/*

Ringback.c V1.3

This program implements a ring-back system that allows a phone line 
that is normally used as a voice line and a dial-out data line to be 
used as a limited dial-in data and voice line.  It does not interfere 
with the use of the line for an answering machine.  

It solves a problem for my situation:

1) I am normally in my office.  The phone line is used for dial-out via
modem and dial-in/dial-out voice traffic.

2) I sometimes travel, at which time I connect an answering machine to
the line.  

3) I'd like to be able to call in to my system via modem when I am 
travelling, without interfering with others use of my answering 
machine.  

4) I don't want to pay for another phone line.

This program functions by watching a tty port for an incoming call on 
a smart modem.  Forty seconds after an incoming call, it will spawn a 
getty on the line.  If no one calls and logs in, the getty will be 
killed by an uncaught alarm and the line will be set back to a no 
answer state.  

This implements a ringback dialin system.  Just call, wait a minute,
and call again. 

If the program notices that a user (uucp or kermit) is using the line 
for an outgoing call, it will back off until that call is completed.  
(Note: the first modem response char will be eaten.  Uucp dialing scripts 
should be set up to tolerate this.) 

Written for Sys V.3 and HDB uucp (but easily modified for others)

Two arguments must be present - tty and type

Sample lines from /etc/inittab:
     rb:1234:respawn:/etc/ringback tty1 1200

Make sure that you trust the people who are allowed to login on this 
line.  It is possible for them to dial-out on the line.  

To install:

1) Edit TTYID and TTYGID and UUGETTY.
2) Compile and install as /etc/ringback
3) Create an initab line like the above one.

 				- Jon Zeeff (umix!b-tech!zeeff)

This program is in the public domain.

*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <termio.h>

char *strcpy();
char *strcat();
unsigned sleep();
unsigned alarm();
void exit();

/* the id and gid for normal ownership of tty lines (usually uucp) */

#define TTYID		5	/* uucp */
#define TTYGID		11	/* staff */

/* if you have a getty/tty/modem combination that doesn't get in a babbling
   getty loop, then there is no need to define this.  If you do, make sure
   that /etc/uugetty exists.
*/ 
#define UUGETTY		/* #define UUGETTY */

main(argc,argv)
int argc;
char **argv;
{

char c;
FILE *tty0,*tty1,*tty2;
#ifdef DEBUG
FILE *console;
#endif
FILE *tty;
FILE *lock;
struct termio ttymodes;
struct stat buf;
char tty_lock[100];
char tty_name[100];
int i;

/* name and location of tty locks - change this as needed */

strcpy(tty_lock,"/usr/spool/locks/LCK.."); 
strcat(tty_lock,argv[1]);

strcpy(tty_name,"/dev/");
strcat(tty_name,argv[1]);

if (argc != 3) exit(2);

unlink(tty_lock);
umask(022);

signal(SIGHUP,SIG_IGN);
signal(SIGINT,SIG_IGN);
signal(SIGQUIT,SIG_IGN);
signal(SIGTERM,SIG_IGN);

for (;;) {
   
      /* Set the line permissions, owner, and group for dial-out use */
   
      chown(tty_name,TTYID,TTYGID);
      chmod(tty_name,0660);
   
      /* Open line - getty will close the first 3, so make sure we are #4 */
      /* so that the modem doesn't get reset on dtr drop and our autoanswer */
      /* setting don't get undone. */

      tty0 = fopen(tty_name,"r+");
      tty1 = fopen(tty_name,"r+");
      tty2 = fopen(tty_name,"r+");
      tty = fopen(tty_name,"r+");

      fclose(tty0); fclose(tty1); fclose(tty2);

#ifdef DEBUG
console = fopen("/dev/console","w");
fprintf(console,"check 1\n"); fflush(console);
#endif

      ttymodes.c_iflag = 0;
      ttymodes.c_oflag = NL1 | CR3;
      ttymodes.c_cflag = 0; 
      ttymodes.c_lflag = 0;
      ttymodes.c_line = 0;
      for (i = 0; i < NCC; ++i) ttymodes.c_cc[i] = 0; 
      ttymodes.c_cc[VMIN] = 1;		/* min number of characters */
      ttymodes.c_cc[VTIME] = 0;     	/* no max time before return */
      ioctl(fileno(tty),TCSETA,&ttymodes);  /* hang up the line */
      sleep(1);
      ttymodes.c_cflag = B1200 | CS8 | CREAD | HUPCL; 
      ioctl(fileno(tty),TCSETA,&ttymodes);
      sleep(1);
#ifdef DEBUG
fprintf(console,"check 1.3\n"); fflush(console);
#endif
   
      /* setup the modem */
   
      fputs("AT\r",tty); fflush(tty); sleep(1);  
      fputs("ATZ\r",tty); fflush(tty); sleep(2);  
      fputs("ATS0=0\r",tty); fflush(tty); sleep(1);  /* Don't answer phone */
      fputs("ATE0\r",tty); fflush(tty); sleep(1);
      fputs("ATV0\r",tty); fflush(tty); sleep(1);
      fputs("ATQ0\r",tty); fflush(tty); sleep(1);
   
#ifdef DEBUG
fprintf(console,"check 2\n"); fflush(console);
#endif
      fcntl(fileno(tty),F_SETFL,O_NDELAY);    /* Set to nonblocking reads */
      while (read(fileno(tty),&c,(unsigned)1) == 1);   /* Flush input */
      fcntl(fileno(tty),F_SETFL,O_RDWR);      /* Set to blocking reads */
   
      /* Wait for incoming call ('2') or other usage (uucp or user) */
   
      /* Get char from tty port */

      while (read(fileno(tty),&c,(unsigned)1) != 1);

      /* 
         If we receive a '2' and no lock file is present, we can be sure
         that it is not someone calling out.
      */   

#ifdef DEBUG
fprintf(console,"check 2.5 %d\n",c); fflush(console);
#endif

      if ((c&0x7f) == '2' && stat(tty_lock,&buf) == -1) {	

	      /* create a hdb style lock file */

              lock = fopen(tty_lock,"w");
      	      if (lock) {
      	         fprintf(lock,"%10.10d\n",(int)getpid());
      	         fclose(lock);
   	      }

	      /* wait for this call to end */
     
              sleep(32);    /* should be long enough for answering machine
			       to answer if they let it ring. */
  
	      /* Setup modem for incoming call */
 
              fputs("AT\r",tty);fflush(tty);sleep(1);
              fputs("ATZ\r",tty);fflush(tty);sleep(1);
              fputs("ATQ1\r",tty);fflush(tty);sleep(1); 
              fputs("ATS0=1\r",tty);fflush(tty);sleep(5);

#ifdef DEBUG
fprintf(console,"check 4\n"); fflush(console);
#endif

      	      /* set things up so the getty will live a short life */
 
              alarm(120);
 
#ifndef UUGETTY  
              execl("/etc/getty","getty","-t","90","-h",
                   argv[1],argv[2],(char *)0);
#else
              unlink(tty_lock);
              execl("/etc/uugetty","uugetty","-r","-h","-t","90",
                   argv[1],argv[2],(char *)0);
#endif

#ifdef DEBUG
fprintf(console,"check 5\n"); fflush(console);
#endif
   	} else {                         /* Must be someone using line */
   
              fclose(tty);                 /* Release line */
   
              /* Loop until lock file goes away (uucp or user done) */
    
              do {
                 while (stat(tty_lock,&buf) == 0) sleep(40);
   	         sleep(40);
   	      }  while (stat(tty_lock,&buf) == 0);
   
   	} /* end_if */
   
   } /* next */
   
}
